/*
 *  - common/d64.c
 *
 *  - disk image (.D64) handling functions
 *  - supported are only 35 Track images without error info (174848 bytes)
 *
 *  This file is part of the VC 1541 Emulator.
 *  Copyright (c) 1997 Torsten Paul
 */

#include <stdio.h>

#include "fileio.h"

extern int timeout; 			           /* this is set by Wait */

static int d64_check_type(char *path, struct stat *st)
{
	int a;

	if (!(st->st_mode & S_IFREG)) return 0;
	if (st->st_size != 174848) return 0;

	return 1;
}

static int d64_read_sector(byte_t *buf, int track, int sector, FILE *f)
{
	long a, offset;
	int stab[] = {
		21, 21, 21, 21, 21, 21, 21, 21, 21, 21,
		21, 21, 21, 21, 21, 21, 21,
		19, 19, 19, 19, 19, 19, 19,
		18, 18, 18, 18, 18, 18,
		17, 17, 17, 17, 17
	};

	if (track <= 0 || track > 35) return 0;
	if (sector < 0 || sector >= stab[track - 1]) return 0;
	offset = 0;
	for (a = 0;a < track - 1;a++) offset += stab[a];
	offset = (offset + sector) * 256;
	if (fseek(f, offset, SEEK_SET) != 0) return 0;
	if (fread(buf, 256, 1, f) != 1) return 0;

	return 1;
}

static int d64_free_blocks(unsigned char *buf)
{
	int a, b;

	/*
	 *  count free block in BAM but
	 *  ignore the directory track 18
	 */
	b = 0;
	for (a = 4;a < 143;a += 4) {
		if (a != 72) b += buf[a];
	}

	return b;
}

static int d64_find_file(char *filename, FILE *infile, char *filename_ret,
			byte_t *track, byte_t *sector, word_t *blocks)
{
	int a, b;
	char fname[17];
	byte_t buf[256];

	d64_read_sector(buf, 18, 0, infile);
	do {
		d64_read_sector(buf, buf[0], buf[1], infile);
		for (a = 0;a < 256;a += 32) {
			/*
			 *  ignore everything else than normal prg file
			 *  (but it may be protected
			 */
			if ((buf[a + 2] & 0xbf) == 0x82) {
				for (b = 0;b < 16 && buf[a + b + 5] != 0xa0;b++) {
					fname[b] = buf[a + b + 5];
				}
				fname[b] = '\0';
				if (util_glob_file(filename, fname, 0)) {
					*track = buf[a + 3];
					*sector = buf[a + 4];
					*blocks = buf[a + 30] + 256 * buf[a + 31];
					strcpy(filename_ret, fname);
					return 1;
				}
			}
		}
	} while (buf[0] != 0);
	return 0;
}

static int d64_send_dir(device_t *d)
{
	FILE *f;
	char *type;
	byte_t buf[256];
	word_t addr;
	word_t a, b, c, e, blocks, free;

	if ((f = fopen(d->path, "rb")) == NULL) return 0;

	if (!d64_read_sector(buf, 18, 0, f)) return 0;

	addr = 0x0801;
	PutIECByte(addr, 0);
	PutIECByte(addr >> 8, 0);
	if (timeout) return 0;
	addr += 30;
	PutIECByte(addr, 0);
	PutIECByte(addr >> 8, 0);
	PutIECByte(0x00, 0);				 /* line number 0 */
	PutIECByte(0x00, 0);
	PutIECByte(0x12, 0);				        /* invers */
	PutIECByte('"', 0);
	if (timeout) return 0;
	for (a = 0;a < 16;a++) {
		PutIECByte(buf[144 + a], 0);
		if (timeout) return 0;
	}
	PutIECByte('"', 0);
	PutIECByte(' ', 0);
	PutIECByte(buf[162], 0);			       /* disk id */
	PutIECByte(buf[163], 0);			       /* disk id */
	PutIECByte(buf[164] & 0x7f, 0);	  	         /* shifted space */
	PutIECByte(buf[165], 0);	           /* dos version/ format */
	PutIECByte(buf[166], 0);		   /* dos version/ format */
	PutIECByte(0x00, 0);				   /* end of line */

	free = d64_free_blocks(buf);

	do {
		if (buf[0] != 18) return 0;
		if (!d64_read_sector(buf, buf[0], buf[1], f)) return 0;
		for (a = 0;a < 256;a += 32) {
			if ((buf[a + 2] & 0x0f) == 0) continue;
			blocks = buf[a + 30] + 256 * buf[a + 31];
			b = 1;
			if (blocks < 100) b++;
			if (blocks < 10)  b++;
			addr += 28 + b;
			if (buf[a + 2] & 0x40) addr++;
			PutIECByte(addr, 0);	       /* next basic line */
			PutIECByte(addr >> 8, 0);
			PutIECByte(buf[a + 30], 0);	    /* blocks low */
			PutIECByte(buf[a + 31], 0);	   /* blocks high */
			if (timeout) return 0;
			for (c = 0;c < b;c++) PutIECByte(' ', 0);
			PutIECByte('"', 0);
			if (timeout) return 0;
			e = 0;
			for (c = 0;c < 16;c++) {
				if (buf[a + 5 + c] == 0xa0 && e == 0) {
					e = 1;
				}
				switch (e) {
				case 0:
					PutIECByte(buf[a + 5 + c], 0);
					break;
				case 1:
					PutIECByte('"', 0);
					e = 2;
					break;
				case 2: PutIECByte(' ', 0);
					break;
				}
				if (timeout) return 0;
			}
			if (e == 0) {
				PutIECByte('"', 0);
			} else {
				PutIECByte(' ', 0);
			}
			PutIECByte(' ', 0);
			if (timeout) return 0;
			/*
			 *  deleted file ?
			 */
			if (buf[a + 2] & 0x80) {
				PutIECByte(' ', 0);
			} else {
				PutIECByte('*', 0);
			}
			switch (buf[a + 2] & 0x0f) {
			case 1:  type = "SEQ"; break;
			case 2:  type = "PRG"; break;
			case 3:  type = "USR"; break;
			case 4:  type = "REL"; break;
			default: type = "???"; break;
			}
			for (c = 0;c < 3;c++) PutIECByte(type[c], 0);
			/*
			 *  protected file ?
			 */
			if (buf[a + 2] & 0x40) {
				PutIECByte('<', 0);
			}

			PutIECByte(0x00, 0);
			if (timeout) return 0;
		}
	} while (buf[0] != 0);

	addr += 16;
	PutIECByte(addr, 0);	 		       /* next basic line */
	PutIECByte(addr >> 8, 0);
	PutIECByte(free, 0);				   /* blocks free */
	PutIECByte(free >> 8, 0);
	if (timeout) return 0;
	for (a = 0;a < 12;a++) {
		PutIECByte("BLOCKS FREE."[a], 0);
		if (timeout) return 0;
	}

	fclose(f);

	PutIECByte(0x00, 0);
	PutIECByte(0x00, 0);
	PutIECByte(0x00, 1);

	return 1;
}

static int d64_send_file(device_t *d, char *filename)
{
	FILE *f;
	char fname[17];
	byte_t buf[256];
	int a, c, blocks;

	if ((f = fopen(d->path, "rb")) == NULL) return 0;

	c = 0;
	if (d64_find_file(filename, f, fname, &buf[0], &buf[1], &blocks)) {
		do {
			c++;
			FileInfo(d->dev, "[%03d/ %03d: T%02d S%02d] '%s'", c,
					blocks, buf[0], buf[1], fname);
			if (!d64_read_sector(buf, buf[0], buf[1], f)) return 0;
			if (buf[0] == 0) {
				/*
				 *  last block
				 */
				for (a = 2;a <= buf[1];a++) {
					PutIECByte(buf[a], 0);
					if (timeout) return 0;
				}
				fclose(f);
				PutIECByte(buf[a], 1);
				return 1;
			} else {
				for (a = 2;a < 256;a++) {
					PutIECByte(buf[a], 0);
					if (timeout) return 0;
				}
			}
		} while (buf[0] != 0);
	}

	return 0;
}

static file_ops_t d64_fileops =
{
	"D64",          /* name */
	d64_check_type, /* check_type */
	d64_send_dir,   /* send_dir */
	d64_send_file,  /* send_file */
};

void register_d64(void)
{
	fileio_register(&d64_fileops);
}
