/*
 *  - common/fileio.c
 *
 *  - generic file and device io functions
 *
 *  This file is part of the VC 1541 Emulator.
 *  Copyright (c) 1997 Torsten Paul
 */

#include <stdio.h>
#include <fcntl.h>
#include <dirent.h>

#include "fileio.h"

static file_ops_t *file_ops[MAX_FILE_OPS];
static device_t devices[MAX_DEVICES];

#ifdef MSDOS
static void interrupt (*old24handler)(void);
#endif /* MSDOS */

void fileio_init(void)
{
	int a;

	for (a = 0;a < MAX_FILE_OPS;a++) {
		file_ops[a] = NULL;
	}
}

void fileio_register(file_ops_t *ops)
{
	int a;

	for (a = 0;a < MAX_FILE_OPS;a++) {
		if (file_ops[a] == NULL) {
			file_ops[a] = ops;
			return;
		}
	}
}

file_ops_t * fileio_check_type(char *path, struct stat *st)
{
	int a;

	for (a = 0;a < MAX_FILE_OPS;a++) {
		if (file_ops[a] == NULL) return NULL;
		if (file_ops[a]->check_type(path, st)) {
			return file_ops[a];
		}
	}
	return NULL;
}

#ifdef MSDOS
void far int24handler(void)
{
		asm	pop	bp
		asm	sti
		asm	mov	al, 0x03
		asm	iret
}

int check_device(int device)
{
	int a, curdisk;

	curdisk = getdisk();
	setdisk(device);
	a = (device == getdisk());
	setdisk(curdisk);
	if (!a) return 0;

	/*
	 *  special handling for a: and b:
	 */
	if ((device == 0) || (device == 1)) {
		asm	int	0x11	       /* BIOS: get configuration */
		asm	mov	a, ax

		if ((a & 1) == 0) return 0;
		a = (a >> 6) & 3;
		if (device > a) return 0;
	}

	return 1;
}
#endif /* MSDOS */

void device_init(void)
{
	int a;

	for (a = 0;a < MAX_DEVICES;a++) {
		strcpy(devices[a].path, DEVICE_PATH_INIT);
		devices[a].dev = FIRST_DEVICE + a;
		devices[a].file_ops = NULL;
	}

#ifdef MSDOS
	setvect(0x24, int24handler);
#endif

	return;
}

void device_activate(byte_t dev)
{
	int a;

	for (a = 0;a < MAX_DEVICES;a++) {
		if (devices[a].dev == dev) {
			device_chdir(dev, "");
			return;
		}
	}
}

int device_isactive(byte_t dev)
{
	int a;

	for (a = 0;a < MAX_DEVICES;a++) {
		if (devices[a].dev == dev) {
			if (devices[a].file_ops != NULL) return 1;
			return 0;
		}
	}
	return 0;
}

int device_chdir(byte_t dev, char *filename)
{
	int a;
	FILE *f;
	device_t *d;
	struct stat st;
	char path[PATH_MAX];
	file_ops_t *file_ops;

	d = NULL;
	for (a = 0;a < MAX_DEVICES;a++) {
		if (devices[a].dev == dev) {
			d = &devices[a];
			break;
		}
	}
	if (d == NULL) return 0;

	if (!device_isactive(dev)) {
		DeviceInfo(d->dev, "[%2d] disabled", d->dev);
	} else {
		DeviceInfo(d->dev, "[%2d] %s: '%s'", d->dev,
				d->file_ops->name, d->path);
	}
	if (filename == NULL) return 0;

#ifdef MSDOS
	for (a = 0;a < strlen(filename);a++) {
		if (filename[a] == '/') filename[a] = '\\';
	}
#endif /* MSDOS */

	if (filename[0] == DIR_SEPARATOR) {
		/*
		 *  absolute path
		 *  /...
		 */
#ifdef MSDOS
		sprintf(path, "%c:%s", d->path[0], filename);
#endif /* MSDOS */
#ifdef LINUX
		strcpy(path, filename);
#endif /* LINUX */
#ifdef MSDOS
	} else if (filename[1] == ':') {
		if (filename[2] != '\\') return 0;
		/*
		 *  absolute path with device letter
		 *  X:/...
		 */
		 strcpy(path, filename);
#endif /* MSDOS */
	} else {
		/*
		 *  relative path
		 */
		util_glob_dir(path, d->path, filename);
	}

#ifdef MSDOS
	if (!check_device(toupper(path[0]) - 'A')) return 0;
#endif /* MSDOS */

	if (stat(path, &st) == -1) return 0;
	if ((file_ops = fileio_check_type(path, &st)) == NULL) return 0;

	strcpy(d->path, path);
	d->file_ops = file_ops;

	DeviceInfo(d->dev, "[%2d] %s: '%s'", d->dev,
					d->file_ops->name, d->path);

	return 1;
}

int device_dir(byte_t dev)
{
	int a;
	device_t *d;

	d = NULL;
	for (a = 0;a < MAX_DEVICES;a++) {
		if (devices[a].dev == dev) {
			d = &devices[a];
			break;
		}
	}
	if (d == NULL) return 0;

	FileInfo(dev, "directory");

	return d->file_ops->send_dir(d);
}

int device_open(byte_t dev, char *filename)
{
	int a;
	device_t *d;

	d = NULL;
	for (a = 0;a < MAX_DEVICES;a++) {
		if (devices[a].dev == dev) {
			d = &devices[a];
			break;
		}
	}
	if (d == NULL) return 0;

	return d->file_ops->send_file(d, filename);
}
