/*
 *  - common/lnx.c
 *
 *  - functions for handling archive files created by the LYNX archiver (.LNX)
 *
 *  This file is part of the VC 1541 Emulator.
 *  Copyright (c) 1997 Torsten Paul
 */

#include <stdio.h>

#include "fileio.h"

extern int timeout; 			           /* this is set by Wait */

#define BLOCK_LEN 254
#define CHECK_LEN BLOCK_LEN

static int lnx_check_type(char *path, struct stat *st)
{
	FILE *f;
	int a, b;
	byte_t buf[CHECK_LEN], *ptr;

	if (!(st->st_mode & S_IFREG)) return 0;
	if ((f = fopen(path, "rb")) == NULL) return 0;

	a = fread(buf, 1, CHECK_LEN, f);
	fclose(f);

	if (a != CHECK_LEN) return 0;

	for (a = 0;a < CHECK_LEN - 2;a++) {
		if ((buf[a] | buf[a + 1] | buf[a + 2]) == 0) break;
	}
	if ((a += 3) >= CHECK_LEN) return 0;
	if (buf[a++] != 0x0d) return 0;

	while ((a < CHECK_LEN) && buf[a] == 0x20) a++;
	while ((a < CHECK_LEN) && buf[a] != 0x20) a++;
	while ((a < CHECK_LEN) && buf[a] == 0x20) a++;
	b = a;
	while ((b < CHECK_LEN) && buf[b] != 0x0d) b++;
	if (b >= CHECK_LEN) return 0;

	ptr = &buf[a];
	buf[b] = '\0';

	if (strstr(ptr, "LYNX") == 0) return 0;

	return 1;
}

static int lnx_get_header(byte_t *buf, word_t *dirlen, word_t *direntries)
{
	byte_t *ptr;
	word_t a, b;

	for (a = 0;a < BLOCK_LEN - 2;a++) {
		if ((buf[a] | buf[a + 1] | buf[a + 2]) == 0) break;
	}
	if ((a += 3) >= BLOCK_LEN) return 0;
	if (buf[a++] != 0x0d) return 0;

	/*
	 *  # of dir blocks
	 */
	while ((a < BLOCK_LEN) && (buf[a] == 0x20)) a++;
	b = a;
	while ((b < BLOCK_LEN) && (buf[b] != 0x20)) b++;
	if (b >= BLOCK_LEN) return 0;
	ptr = &buf[a];
	buf[b] = '\0';
	if ((*dirlen = atoi(ptr)) == 0) return 0;
	*dirlen *= BLOCK_LEN;

	/*
	 *  skip over signature
	 */
	while ((a < BLOCK_LEN) && (buf[a] != 0x0d)) a++;
	a++;
	if (a >= BLOCK_LEN) return 0;

	/*
	 *  # of dir entries
	 */
	b = a;
	while ((b < BLOCK_LEN) && (buf[b] != 0x0d)) b++;
	if (b >= BLOCK_LEN) return 0;
	ptr = &buf[a];
	buf[b] = '\0';
	if ((*direntries = atoi(ptr)) == 0) return 0;

	return b + 1;
}

static char *lnx_getstr(word_t *a, byte_t *dirmem, word_t dirlen)
{
	word_t b;
	byte_t *ptr;

	b = *a;
	while ((b < dirlen) && (dirmem[b] != 0x0d)) b++;
	if (b >= dirlen) return NULL;
	ptr = &dirmem[*a];
	dirmem[b] = '\0';
	*a = b + 1;

	return ptr;
}

static int lnx_send_dir(device_t *d)
{
	FILE *f;
	word_t addr;
	word_t a, b, c;
	word_t dirlen, direntries;
	byte_t *name, *len, *type, *last;
	byte_t buf[BLOCK_LEN], *ptr, *dirmem;
	word_t blocks;

	if ((f = fopen(d->path, "rb")) == NULL) return 0;

	if (fread(buf, 1, BLOCK_LEN, f) != BLOCK_LEN) {
		fclose(f);
		return 0;
	}

	if ((c = lnx_get_header(buf, &dirlen, &direntries)) == 0) return 0;

	rewind(f);
	if ((dirmem = (byte_t *)malloc(dirlen)) == NULL) {
		return 0;
	}
	if (fread(dirmem, 1, dirlen, f) != dirlen) {
		fclose(f);
		return 0;
	}

	addr = 0x0801;
	PutIECByte(addr, 0);
	PutIECByte(addr >> 8, 0);
	if (timeout) return 0;
	addr += 30;
	PutIECByte(addr, 0);
	PutIECByte(addr >> 8, 0);
	PutIECByte(0x00, 0);				 /* line number 0 */
	PutIECByte(0x00, 0);
	PutIECByte(0x12, 0);				        /* invers */
	PutIECByte('"', 0);
	if (timeout) return 0;
	for (a = 0;a < strlen(d->path);a++) {
		b = toupper(d->path[a]);
		if (b == '\\') b = '/';
		PutIECByte(b, 0);
		if (timeout) return 0;
	}
	PutIECByte('"', 0);
	PutIECByte(0x00, 0);				   /* end of line */
	a = c;
	for (c = 0;c < direntries;c++) {
		if ((name = lnx_getstr(&a, dirmem, dirlen)) == NULL) return 0;
		if ((len  = lnx_getstr(&a, dirmem, dirlen)) == NULL) return 0;
		if ((type = lnx_getstr(&a, dirmem, dirlen)) == NULL) return 0;
		if ((last = lnx_getstr(&a, dirmem, dirlen)) == NULL) return 0;

		switch (*type) {
		case 'P': type = "PRG"; break;
		case 'S': type = "SEQ"; break;
		case 'R': type = "REL"; break;
		case 'U': type = "USR"; break;
		default:  type = "???"; break;
		}

		blocks = atoi(len);
		if (util_send_filename(&addr, blocks, name, type) == 0) {
			return 0;
		}
	}

	PutIECByte(0x00, 0);
	PutIECByte(0x00, 1);

	return 1;
}

static int lnx_send_file(device_t *d, char *filename)
{
	FILE *f;
	word_t a, b, c, found;
	word_t dirlen, direntries, blocks;
	byte_t *name, *len, *type, *last;
	byte_t buf[BLOCK_LEN], *ptr, *dirmem, tmp;

	if ((f = fopen(d->path, "rb")) == NULL) return 0;

	if (fread(buf, 1, BLOCK_LEN, f) != BLOCK_LEN) {
		fclose(f);
		return 0;
	}

	if ((a = lnx_get_header(buf, &dirlen, &direntries)) == 0) return 0;

	rewind(f);
	if ((dirmem = (byte_t *)malloc(dirlen)) == NULL) {
		return 0;
	}
	if (fread(dirmem, 1, dirlen, f) != dirlen) {
		fclose(f);
		return 0;
	}

	found = 0;
	blocks = 0;
	for (b = 0;b < direntries;b++) {
		if ((name = lnx_getstr(&a, dirmem, dirlen)) == NULL) return 0;
		if ((len  = lnx_getstr(&a, dirmem, dirlen)) == NULL) return 0;
		if ((type = lnx_getstr(&a, dirmem, dirlen)) == NULL) return 0;
		if ((last = lnx_getstr(&a, dirmem, dirlen)) == NULL) return 0;
		if (util_glob_file(filename, name, 0)) {
			found = 1;
			break;
		}
		blocks += atoi(len);
	}
	if (found) {
		if (fseek(f, blocks * BLOCK_LEN + dirlen, SEEK_SET) != 0) {
			fclose(f);
			return 0;
		}
		blocks = atoi(len);
		b = c = 0;
		for (a = 0;a < (blocks - 1) * BLOCK_LEN;a++) {
			PutIECByte(fgetc(f), 0);
			if (timeout) return 0;
			if ((b++ % 254) == 0) {
				FileInfo(d->dev, "[%03d/ %03d] '%s'",
						c++, blocks, name);
			}
		}
		for (a = 0;a < atoi(last) - 2;a++) {
			PutIECByte(fgetc(f), 0);
			if (timeout) return 0;
			if ((b++ % 254) == 0) {
				FileInfo(d->dev, "[%03d/ %03d] '%s'",
						c++, blocks, name);
			}
		}
		FileInfo(d->dev, "[%03d/ %03d] '%s'", c, blocks, name);

		tmp = fgetc(f);
		fclose(f);
		PutIECByte(tmp, 1);
		if (timeout) return 0;
	}

	return 1;
}

static file_ops_t lnx_fileops =
{
	"LNX",          /* name */
	lnx_check_type, /* check_type */
	lnx_send_dir,   /* send_dir */
	lnx_send_file,  /* send_file */
};

void register_lnx(void)
{
	fileio_register(&lnx_fileops);
}
