/*
 *  - common/t64.c
 *
 *  - tape image (.T64) handling functions
 *
 *  This file is part of the VC 1541 Emulator.
 *  Copyright (c) 1997 Torsten Paul
 */

#include <stdio.h>

#include "fileio.h"
#include "t64.h"

#define CHECK_LEN 20

extern int timeout; 			           /* this is set by Wait */

static int t64_check_type(char *path, struct stat *st)
{
	int a;
	FILE *f;
	char buf[CHECK_LEN];

	if (!(st->st_mode & S_IFREG)) return 0;
	if ((f = fopen(path, "rb")) == NULL) return 0;

	a = fread(buf, 1, CHECK_LEN, f);
	fclose(f);

	if (a != CHECK_LEN) return 0;

	if (strncmp(buf, "C64 tape file", 13) == 0) return 1;
	if (strncmp(buf, "C64 tape image file", 19) == 0) return 1;
	if (strncmp(buf, "C64S tape file", 14) == 0) return 1;
	if (strncmp(buf, "C64S tape image file", 20) == 0) return 1;

	return 0;
}

static int t64_send_dir(device_t *d)
{
	int a, b, c;
	FILE *f;
	char *type;
	t64_tape_t tr;
	t64_file_t fr;
	word_t addr, blocks;

	if ((f = fopen(d->path, "rb")) == NULL) return 0;

	addr = 0x0801;
	PutIECByte(addr, 0);
	PutIECByte(addr >> 8, 0);
	if (timeout) return 0;
	addr += 32;
	PutIECByte(addr, 0);
	PutIECByte(addr >> 8, 0);
	PutIECByte(0x00, 0);		/* line number 0 */
	PutIECByte(0x00, 0);
	PutIECByte(0x12, 0);		/* invers */
	PutIECByte('"', 0);
	if (timeout) return 0;
	if (fread(&tr, sizeof(t64_tape_t), 1, f) != 1) {
		fclose(f);
		return 0;
	}
	for (a = 0;a < 24;a++) {
		PutIECByte(tr.user_desc[a], 0);
		if (timeout) return 0;
	}
	PutIECByte('"', 0);
	PutIECByte(0x00, 0);
	for (a = 0;a < tr.entries;a++) {
		if (fread(&fr, sizeof(t64_file_t), 1, f) != 1) {
			fclose(f);
			return 0;
		}
		if (fr.entry_type != 0) {
			switch (fr.entry_type) {
			case 1:  type = "FILE"; break;
			case 3:  type = "SNAPSHOT V0.9"; break;
			default: type = "???"; break;
			}
			b = 1;
			blocks = (fr.end - fr.start + 253) / 254;
			if (blocks < 100) b++;
			if (blocks < 10) b++;
			addr += 25 + b + strlen(type);
			PutIECByte(addr, 0);		/* next basic line */
			PutIECByte(addr >> 8, 0);
			PutIECByte(blocks, 0);		/* blocks low */
			PutIECByte(blocks >> 8, 0);	/* blocks high */
			if (timeout) return 0;
			for (c = 0;c < b;c++) PutIECByte(' ', 0);
			PutIECByte('"', 0);
			if (timeout) return 0;
			for (c = 0;c < 16;c++) PutIECByte(fr.name[c], 0);
			PutIECByte('"', 0);
			PutIECByte(' ', 0);
			PutIECByte(' ', 0);
			if (timeout) return 0;
			for (c = 0;c < strlen(type);c++) PutIECByte(type[c], 0);
			PutIECByte(0x00, 0);
			if (timeout) return 0;
		}
	}
	fclose(f);

	PutIECByte(0x00, 0);
	PutIECByte(0x00, 1);

	return 0;
}

static int t64_send_file(device_t *d, char *filename)
{
	FILE *f;
	int found;
	byte_t tmp;
	char fname[17];
	t64_tape_t tr;
	t64_file_t fr;
	word_t a, b, c, blocks;

	if ((f = fopen(d->path, "rb")) == NULL) return 0;

	if (fread(&tr, sizeof(t64_tape_t), 1, f) != 1) {
		fclose(f);
		return 0;
	}

	found = 0;
	for (a = 0;a < tr.entries;a++) {
		if (fread(&fr, sizeof(t64_file_t), 1, f) != 1) {
			fclose(f);
			return 0;
		}
		if (fr.entry_type != 0) {
			for (b = 0;b < 16;b++) fname[b] = fr.name[b];
			fname[b] = '\0';
		}
		if (util_glob_file(filename, fname, 0)) {
			found = 1;
			break;
		}
	}
	if (found) {
		blocks = (fr.end - fr.start + 253) / 254;

		PutIECByte(fr.start, 0);
		PutIECByte(fr.start >> 8, 0);
		if (fseek(f, fr.offset, SEEK_SET) != 0) {
			fclose(f);
			return 0;
		}
		b = c = 0;
		for (a = fr.start;a < fr.end - 1;a++) {
			PutIECByte(fgetc(f), 0);
			if (timeout) return 0;
			if ((b++ % 254) == 0) {
				FileInfo(d->dev, "[%03d/ %03d] '%s'",
						c++, blocks, fname);
			}
		}
		FileInfo(d->dev, "[%03d/ %03d] '%s'", c, blocks, fname);

		tmp = fgetc(f);
		fclose(f);
		PutIECByte(tmp, 1);
		if (timeout) return 0;
	} else {
		fclose(f);
	}


	return found;
}

static file_ops_t t64_fileops =
{
	"T64",          /* name */
	t64_check_type, /* check_type */
	t64_send_dir,   /* send_dir */
	t64_send_file,  /* send_file */
};

void register_t64(void)
{
	fileio_register(&t64_fileops);
}

